@@ -2,7 +2,7 @@
 //
 // RcppEigenWrap.h: Rcpp wrap methods for Eigen matrices, vectors and arrays
 //
-// Copyright (C) 2011 - 2012   Douglas Bates, Dirk Eddelbuettel and Romain Francois
+// Copyright (C) 2011 - 2022   Douglas Bates, Dirk Eddelbuettel and Romain Francois
 //
 // This file is part of RcppEigen.
 //
@@ -80,16 +80,22 @@ namespace Rcpp{
 
         // for plain dense objects
         template <typename T>
-        SEXP eigen_wrap_plain_dense( const T& obj, Rcpp::traits::true_type ){
-			typename Eigen::internal::conditional<T::IsRowMajor,
-												  Eigen::Matrix<typename T::Scalar,
-																T::RowsAtCompileTime,
-																T::ColsAtCompileTime>,
-												  const T&>::type objCopy(obj);
-            int m = obj.rows(), n = obj.cols();
-			R_xlen_t size = static_cast<R_xlen_t>(m) * n;
-			SEXP ans = PROTECT(::Rcpp::wrap(objCopy.data(), objCopy.data() + size));
-            if( T::ColsAtCompileTime != 1 ) {
+        SEXP eigen_wrap_plain_dense( const T& obj, Rcpp::traits::true_type ) {
+            bool needs_dim = T::ColsAtCompileTime != 1;
+            R_xlen_t m = obj.rows(), n = obj.cols();
+            if (needs_dim && (m > INT_MAX || n > INT_MAX)) {
+                Rcpp::stop("array dimensions cannot exceed INT_MAX");
+            }
+            R_xlen_t size = m * n;
+            typename Eigen::internal::conditional<
+                T::IsRowMajor,
+                Eigen::Matrix<typename T::Scalar,
+                              T::RowsAtCompileTime,
+                              T::ColsAtCompileTime>,
+                const T&>::type objCopy(obj);
+            SEXP ans = PROTECT(::Rcpp::wrap(objCopy.data(),
+                                            objCopy.data() + size));
+            if (needs_dim) {
                 SEXP dd = PROTECT(::Rf_allocVector(INTSXP, 2));
                 int *d = INTEGER(dd);
                 d[0] = m;
@@ -213,7 +219,7 @@ namespace Rcpp{
         public:
             Exporter(SEXP x) : vec(x) {
                 if (TYPEOF(x) != RTYPE)
-                    throw std::invalid_argument("Wrong R type for mapped vector");
+                    throw std::invalid_argument("Wrong R type for mapped vector"); // #nocov
             }
             OUT get() {return OUT(vec.begin(), vec.size());}
         } ;
@@ -257,7 +263,7 @@ namespace Rcpp{
             public:
             Exporter(SEXP x) : vec(x), d_ncol(1), d_nrow(Rf_xlength(x)) {
                 if (TYPEOF(x) != RTYPE)
-                    throw std::invalid_argument("Wrong R type for mapped matrix");
+                    throw std::invalid_argument("Wrong R type for mapped matrix");	// #nocov
                 if (::Rf_isMatrix(x)) {
                     int *dims = INTEGER( ::Rf_getAttrib( x, R_DimSymbol ) ) ;
                     d_nrow = dims[0];
